/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package engcalculator.function.infix;

import engcalculator.function.infix.interval.*;
import engcalculator.function.infix.list.*;
import engcalculator.function.infix.engineering.*;
import engcalculator.function.infix.matrix.*;
import engcalculator.function.infix.set.*;
import engcalculator.function.infix.arithmetic.*;
import engcalculator.function.Function;
import engcalculator.function.infix.compare.*;
import engcalculator.function.infix.logic.*;
import engcalculator.data.IntervalList;
import engcalculator.data.MatrixIntervalList;
import engcalculator.expression.Expression;
import engcalculator.expression.token.InfixFunctionTkn;
import engcalculator.function.CalculusException;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;


/**
 *
 * @author simone
 */
public abstract class InfixFunction extends Function {
    private static HashMap<String, InfixFunction> INFIX_FUNCTIONS = new HashMap <String, InfixFunction> ();
    private static LinkedList<Boolean> ASSOCIATIVITY = new LinkedList<Boolean> ();
    private final static InfixFunctionTkn INFIX_FUNCTION_TKN = new InfixFunctionTkn();
    private double MAX_CHECK_ERROR = 1e-3;

    public static boolean addFunction (InfixFunction func) throws Exception {
        if(! INFIX_FUNCTION_TKN.mayBeCompatible(func.getSymbol())) throw new CalculusException("The infix function name is not properly formed: "+func.getSymbol());

        InfixFunction f;
        if ((f = INFIX_FUNCTIONS.get(func.getSymbol())) != null && f.isConstant()) return false;

        func.priority = getMaxPriority();

        INFIX_FUNCTIONS.put(func.getSymbol(), func);

        return true;
    }

    public static boolean delFunction(String value) {
        Function f;
        if ((f = INFIX_FUNCTIONS.get(value)) != null && ! f.isConstant()) {
            INFIX_FUNCTIONS.remove(value);
            return true;
        }
        else return false;
    }

    public static InfixFunction getFunction (String symbol) {
        return INFIX_FUNCTIONS.get(symbol);
    }

    public static Collection<InfixFunction> getInfixFunctions () {
        return INFIX_FUNCTIONS.values();
    }

    public static byte getMaxPriority () {
        return (byte) (ASSOCIATIVITY.size()-1);
    }

    public static void addPriority (boolean isLeftAssociative) {
        ASSOCIATIVITY.add(isLeftAssociative);
    }

    public static boolean isLeftAssociative (int priority) {
        return ASSOCIATIVITY.get(priority);
    }

    private static String buildHelp (String name, boolean expandsToList, String basehelp, String[] examplesleft, String[] examplesright, String[] results) {
        StringBuffer help = new StringBuffer (basehelp);
        if (expandsToList) help.append(" If the arguments are lists they must have the same number of arguments. If one element is a list and the other is a scalar, the scalar will be automatically expanded to a list with the proper number of elements. The result is the list of the results.");
        if (examplesleft != null && examplesright != null && results != null && examplesleft.length == results.length && examplesright.length == results.length)  {
            help.append("\nExamples: ");
            for (int i=0; i<examplesleft.length; ++i)
                help.append ("\n").append(examplesleft[i]).append(name).append(examplesright[i]).append(" = ").append(results[i]);
        }
        return help.toString();
    }

    public static void initialize () throws Exception {
        boolean noDuplicateName = true;

        addPriority(true);
        noDuplicateName &= addFunction(new ListBuilder());
        noDuplicateName &= addFunction(new MatrixByColumnBuilder());

        addPriority(false);
        noDuplicateName &= addFunction(new Set());

        addPriority(true);
        noDuplicateName &= addFunction(new Or());

        addPriority(true);
        noDuplicateName &= addFunction(new And());

        addPriority(true);
        noDuplicateName &= addFunction(new Union());

        addPriority(true);
        noDuplicateName &= addFunction(new Intersection());
        
        addPriority(true);
        noDuplicateName &= addFunction(new Equal());
        noDuplicateName &= addFunction(new NotEqual());

        addPriority(true);
        noDuplicateName &= addFunction(new Smaller());
        noDuplicateName &= addFunction(new Greater());
        noDuplicateName &= addFunction(new SmallerEqual());
        noDuplicateName &= addFunction(new GreaterEqual());
        noDuplicateName &= addFunction(new SubSet());
        noDuplicateName &= addFunction(new SuperSet());

        addPriority(true);
        noDuplicateName &= addFunction(new Sum());
        noDuplicateName &= addFunction(new Sub());

        addPriority(true);
        noDuplicateName &= addFunction(new Mul());
        noDuplicateName &= addFunction(new Div());
        noDuplicateName &= addFunction(new SumInverted());
        noDuplicateName &= addFunction(new SumSquared());
        noDuplicateName &= addFunction(new MatrixMul());

        addPriority(false);
        noDuplicateName &= addFunction(new Pow());

        addPriority(true);
        noDuplicateName &= addFunction(new Get());
        noDuplicateName &= addFunction(new GetVector());

        addPriority(true);
        noDuplicateName &= addFunction(new IntervalByExtremesBuilder());
        noDuplicateName &= addFunction(new IntervalByToleranceBuilder());
        noDuplicateName &= addFunction(new MatrixBuilder());
        

        if (! noDuplicateName) throw new CalculusException ("Duplicated Infix Function definition while initializing internal functions.");
    }

    private byte numbLeftArgs;          //number of left arguments, <0 for variable
    private byte numbRightArgs;         //number of right arguments, <0 for variable
    private byte priority;              //priority 0 lowest, <0 for infix
    private boolean expandsScalarInput; //if exapndible (and number of arguments is not variable) it is automatically applied to all the components of the list
    private String[] examplesLeft;
    private String[] examplesRight;
    private String[] results;

    public InfixFunction(String name, byte numbLeftArgs, byte numbRightArgs, boolean expandsScalarInput, boolean constant, String help, String[] examplesLeft, String[] examplesRight, String[] results) {
        super(name, constant, buildHelp(name, numbLeftArgs > 0 && numbRightArgs > 0 && expandsScalarInput, help, examplesLeft, examplesRight, results));
        this.numbLeftArgs = numbLeftArgs;
        this.numbRightArgs = numbRightArgs;
        this.expandsScalarInput = expandsScalarInput;
        this.examplesLeft = examplesLeft;
        this.examplesRight = examplesRight;
        this.results = results;
        priority = 0;
    }

    public byte getPriority() {
        return priority;
    }


    @Override
    protected String selfCheck() {
        IntervalList computedRes, expectedRes;
        String expression = null;
        double error;

        if (examplesLeft == null && examplesRight == null && results == null ) return null;
        if (examplesLeft != null && examplesRight != null && results != null && examplesLeft.length == results.length && examplesRight.length == results.length)  {
            for (int i=0; i<examplesLeft.length; ++i) {
                try {
                    expression = new StringBuffer("(").append(examplesLeft[i]).append(") ").append(getSymbol()).append(" (").append(examplesRight[i]).append(")").toString();

                    computedRes = new Expression(expression, null).evaluate(null);
                    expectedRes = new Expression(new StringBuffer("(").append(results[i]).append(")").toString(), null).evaluate(null);

                    if (computedRes.size() != expectedRes.size()) return new StringBuffer(expression).append(" does not match in size with ").append(results[i]).toString();
                    error = 0;
                    for (int j=0; j<computedRes.size(); ++j)
                        error += Math.abs(computedRes.get(j).getRight()-expectedRes.get(j).getRight())+Math.abs(computedRes.get(j).getLeft()-expectedRes.get(j).getLeft());

                } catch (Exception e) {
                    return new StringBuffer("Exception computing: ").append(e.getMessage()).append(expression).toString();
                }
                if (error > MAX_CHECK_ERROR) return new StringBuffer(expression).append(" == ").append(computedRes.toString()).append(" != ").append(results[i]).toString();
            }
        } else return new StringBuffer("Examples size for ").append(getSymbol()).append(" do not match.").toString();
        return null;
    }

    @Override
    public final IntervalList compute (IntervalList... inputArray) throws Exception {
        if (inputArray.length != getNumberOfArguments()) throw new Exception ("It was expected just one argument for a prefix function");
        IntervalList leftSide = inputArray[0],  rightSide = inputArray[1];

        if (! expandsScalarInput && numbLeftArgs > 0 && leftSide.size() != numbLeftArgs) throw new CalculusException ("Wrong number of arguments on the left side of function "+getSymbol());
        if (! expandsScalarInput && numbRightArgs > 0 && rightSide.size() != numbRightArgs) throw new CalculusException ("Wrong number of arguments on the right side of function "+getSymbol());
        if (rightSide.size() == 0 || leftSide.size() == 0) throw new CalculusException ("There is not any operand on one side of function "+getSymbol());
        int columns = -1;
        if (expandsScalarInput) {
            if (leftSide.size()==1 && rightSide.size() > 1) {
                columns = rightSide.getColumns();
                if (rightSide.isMatrix()) leftSide = new MatrixIntervalList (leftSide, columns); //to avoid to affect input
                else leftSide = new IntervalList (leftSide);
                while (leftSide.size() < rightSide.size())
                    leftSide.add(leftSide.getFirst());
            } else if (rightSide.size() == 1 && leftSide.size() > 1) {
                columns = leftSide.getColumns();
                if (leftSide.isMatrix()) rightSide = new MatrixIntervalList (rightSide, columns); //to avoid to affect the input
                else rightSide = new IntervalList (rightSide);
                while (rightSide.size() < leftSide.size())
                    rightSide.add(rightSide.getFirst());
            }
        }

        if (expandsScalarInput && leftSide.size() / numbLeftArgs == rightSide.size() / numbRightArgs &&  leftSide.size() % numbLeftArgs == 0) {
            IntervalList result;
            if (leftSide.isMatrix() && rightSide.isMatrix()) result = new MatrixIntervalList (leftSide.getColumns());
            else result = new IntervalList ();

            int left=0, right=0;
            while (left<leftSide.size())
                result.addAll(_compute(leftSide.subList(left, left+=numbLeftArgs), rightSide.subList(right, right+=numbRightArgs)));
            
            return result;
        } else if ((numbLeftArgs < 0 || leftSide.size() == numbLeftArgs ) && (numbRightArgs < 0 || rightSide.size() == numbRightArgs )) {
            return _compute( leftSide, rightSide);
        } else {
            throw new CalculusException ("Wrong number of arguments evaluating the function "+getSymbol());
        }
    }

    public abstract IntervalList _compute (IntervalList leftSide, IntervalList rightSide) throws Exception;
    
    @Override
    public String getHelp() {
        if (super.getHelp().equals("")) return "";

        StringBuffer result = new StringBuffer();
        result.append("\nName: ").append(getSymbol()).append("\nType: infix function\nNumber of left args: ");
        if (numbLeftArgs < 0) result.append(" variable");
        else result.append(numbLeftArgs);
        result.append("\nNumber of right args: ");
        if (numbRightArgs < 0) result.append(" variable");
        else result.append(numbRightArgs);
        result.append("\nPriority: ").append(priority).append(", Min Priority: 0, Max Priority: ").append(getMaxPriority()).append("\nAssociativity: ");
        if (isLeftAssociative(priority)) result.append("left to right");
        else result.append("right to left");
        if (expandsScalarInput) result.append("\nAutomatically extendes scalar input to list");
        else result.append("\nDoes not extend scalar input to list");
        if (isConstant()) result.append("\nCannot be changed by user");
        else result.append("\nCan be changed by user");
        result.append("\nHelp: ").append(super.getHelp()).append('\n');
        return result.toString();
    }

    @Override
    public final int getNumberOfArguments() {
        return 2;
    }

}
